import operator
import os
import re
import string

from globals import *

rx_thread = re.compile(r'^(\d+):(\S+) \[(\d+)\] (.+)$')
rx_subject = re.compile(r'^(\S+) (.+)$')
rx_message = re.compile(r'^(\d+):(\d+):(\S+) (.+)$')
rx_references = re.compile(r'\nReferences:[^\n]+(\n\s[^\n]+)*\n',
						   re.IGNORECASE)
rx_inreplyto = re.compile(r'\nIn-Reply-To:\s*<([^>]+)>', re.IGNORECASE)
rx_endofhdr = re.compile(r'^\s*$')
rx_messageid = re.compile(r'\nMessage-ID:\s*<([^>]+)>', re.IGNORECASE)
rx_mid = re.compile(r'<([^>]+)>')
rx_ws = re.compile(r'\s+')

def _in(a, min, max):
	if a > max: return max
	if a < min: return min
	return a

month_names = {
	1: 'January',
	2: 'February',
	3: 'March',
	4: 'April',
	5: 'May',
	6: 'June',
	7: 'July',
	8: 'August',
	9: 'September',
	10: 'October',
	11: 'November',
	12: 'December'
	}

def set_months(form):
	list = os.listdir(os.path.join(form[LISTDIR], 'archive', 'threads'))
	list.sort()
	firstmonth = int(list[0])
	lastmonth = int(list[-1])
	try:
		month = int(form[MONTH])
		if month > lastmonth or month < firstmonth:
			raise ValueError, "Month value is out of range."
	except KeyError:
		month = lastmonth
	nextmonth = month + 1
	prevmonth = month - 1
	if nextmonth % 100 > 12: nextmonth = month + 89
	if prevmonth % 100 == 0: prevmonth = prevmonth - 88
	elif prevmonth % 100 > 12: prevmonth = month - 100
	nextmonth = _in(nextmonth, firstmonth, lastmonth)
	prevmonth = _in(prevmonth, firstmonth, lastmonth)
	form[FIRSTMONTH] = firstmonth
	form[PREVMONTH] = prevmonth
	form[MONTH] = month
	form[NEXTMONTH] = nextmonth
	form[LASTMONTH] = lastmonth
	form[MONTHNAME] = month_names[month % 100]
	form[YEAR] = month / 100

def read_threads(form):
	list = [ ]
	path = os.path.join(form[LISTDIR], 'archive', 'threads', str(form[MONTH]))
	for line in open(path).readlines():
		match = rx_thread.match(string.strip(line))
		groups = match.groups()
		list.append({
			MSGNUM: int(groups[0]),
			THREADID: groups[1],
			MSGCOUNT: int(groups[2]),
			SUBJECT: groups[3]
			})
	return list

def read_thread(form):
	threadid = form[THREADID]
	if '.' in threadid:
		raise ValueError, "Thread ID contains invalid characters"
	part1 = threadid[:2]
	part2 = threadid[2:]
	path = os.path.join(form[LISTDIR], 'archive', 'subjects', part1, part2)
	lines = map(string.strip, open(path).readlines())
	subject = rx_subject.match(lines[0]).group(2)
	list = [ ]
	msgnums = [ ]
	for line in lines[1:]:
		match = rx_message.match(line)
		if match:
			groups = match.groups()
			msgnum = int(groups[0])
			list.append({
				MSGNUM: msgnum,
				MONTH: int(groups[1]),
				AUTHORID: groups[2],
				AUTHOR: groups[3],
				})
			msgnums.append(msgnum)
	return (subject, list)

def read_month(form):
	messages = { }
	month = int(form[MONTH])
	for thread in read_threads(form):
		(subject,thread) = read_thread({THREADID: thread[THREADID],
										LISTDIR: form[LISTDIR]})
		for message in thread:
			if message[MONTH] == month:
				message[SUBJECT] = subject
				messages[message[MSGNUM]] = message
	numbers = messages.keys()
	numbers.sort()
	return map(operator.getitem, [messages]*len(numbers), numbers)

def _parse_references(str):
	list = [ ]
	match = rx_mid.search(str)
	while match:
		list.append(match.group(1))
		match = rx_mid.search(str, match.end())
	return list

def _open_message(listdir, num):
	num = '%03d' % num
	return open(os.path.join(listdir, 'archive', num[:-2], num[-2:]))

def open_message(form):
	return _open_message(form[LISTDIR], int(form[MSGNUM]))

def _parse_message(listdir, num):
	# Result: message-id, in-reply-to, references
	headers = _open_message(listdir, num).read(4096)
	eoh = rx_endofhdr.search(headers)
	if eoh: headers = headers[:eoh.start()]
	messageid = rx_messageid.search(headers)
	if messageid: messageid = messageid.group(1)
	inreplyto = rx_inreplyto.search(headers)
	if inreplyto: inreplyto = inreplyto.group(1)
	refs = rx_references.search(headers)
	if refs: refs = _parse_references(refs.group(0))
	return (messageid, inreplyto, refs)

#def _thread_message(id, idmap, nummap, irtmap, replymap, refmap):
def _thread_message(id, idmap, nummap, replymap):
	# result: {replies}
	#try: irt = nummap[irtmap[id]]
	#except KeyError: irt = None
	replies = { }
	for i in replymap[id]:
		num = nummap[i]
		replies[num] = _thread_message(i, idmap, nummap, replymap)
		idmap[num] = None
	#references = [ ]
	#if refmap[id]:
	#	for i in refmap[id]:
	#		if i <> irtmap[id] and nummap.has_key(i):
	#			references.append(nummap[i])
	#followups = [ ]
	#for i in replymap[id]:
	#	if not replymap.has_key(i):
	#		followups.append(nummap[i])
	#return (irt, replies, references, followups)
	return replies

def thread_messages(form, messages):
	listdir = form[LISTDIR]
	messageids = { }
	messagenums = { }
	replies = { }
	inreplytos = { }
	references = { }
	for message in messages:
		msgnum = message[MSGNUM]
		( messageid, inreplyto, refs ) = _parse_message(listdir, msgnum)
		messageids[msgnum] = messageid
		messagenums[messageid] = msgnum
		replies[messageid] = [ ]
		references[messageid] = refs
		if refs and not inreplyto:
			inreplyto = refs[-1]
		if inreplyto:
			if replies.has_key(inreplyto):
				replies[inreplyto].append(messageid)
		inreplytos[messageid] = inreplyto
	#print 'messageids:', messageids
	#print 'messagenums:', messagenums
	#print 'replies:', replies
	#print 'references:', references
	list = { }
	for message in messages:
		msgnum = message[MSGNUM]
		if messageids[msgnum] is not None:
			tmp = _thread_message(messageids[msgnum],
								  messageids, messagenums, replies)
			                      #inreplytos, replies, references)
			list[msgnum] = tmp
	return list

def search_threads(form):
	terms = rx_ws.split(form[TERMS])
	terms = map(string.lower, terms)
	dir = os.path.join(form[LISTDIR], 'archive', 'threads')
	threads = [ ]
	for month in os.listdir(dir):
		for line in open(os.path.join(dir, month)).readlines():
			line = string.strip(line)
			match = rx_thread.match(line)
			if match:
				groups = match.groups()
				subject = string.lower(groups[3])
				for term in terms:
					if string.find(subject, term) < 0: subject = ''; break
				if not subject: continue
				groups = match.groups()
				threads.append({
					MSGNUM: int(groups[0]),
					THREADID: groups[1],
					MSGCOUNT: int(groups[2]),
					SUBJECT: groups[3],
					MONTH: int(month)
					})
	return threads
